import asyncio

from nestipy.microservice import Transport
from nestipy.microservice.client import GrpcClientProxy, MicroserviceOption


async def main():
    client = GrpcClientProxy(
        option=MicroserviceOption(
            transport=Transport.GRPC
        ),
    )
    response = await client.send("test", "request")
    print("Response from server:", response.data)


if __name__ == "__main__":
    asyncio.run(main())



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: microservice.proto
# Protobuf Python Version: 5.28.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    28,
    1,
    '',
    'microservice.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x12microservice.proto\x12\x02ms\"\x1b\n\nSubRequest\x12\r\n\x05topic\x18\x01 \x01(\t\"\x1d\n\x0cUnsubRequest\x12\r\n\x05topic\x18\x01 \x01(\t\"*\n\x0b\x44\x61taRequest\x12\r\n\x05topic\x18\x01 \x01(\t\x12\x0c\n\x04\x64\x61ta\x18\x02 \x01(\t\"+\n\x0c\x44\x61taResponse\x12\x0c\n\x04\x64\x61ta\x18\x01 \x01(\t\x12\r\n\x05topic\x18\x02 \x01(\t\"\x07\n\x05\x45mpty2\x8b\x01\n\x04Grpc\x12/\n\tSubscribe\x12\x0e.ms.SubRequest\x1a\x10.ms.DataResponse0\x01\x12*\n\x0bUnsubscribe\x12\x10.ms.UnsubRequest\x1a\t.ms.Empty\x12&\n\x08SendData\x12\x0f.ms.DataRequest\x1a\t.ms.Emptyb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'microservice_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_SUBREQUEST']._serialized_start=26
  _globals['_SUBREQUEST']._serialized_end=53
  _globals['_UNSUBREQUEST']._serialized_start=55
  _globals['_UNSUBREQUEST']._serialized_end=84
  _globals['_DATAREQUEST']._serialized_start=86
  _globals['_DATAREQUEST']._serialized_end=128
  _globals['_DATARESPONSE']._serialized_start=130
  _globals['_DATARESPONSE']._serialized_end=173
  _globals['_EMPTY']._serialized_start=175
  _globals['_EMPTY']._serialized_end=182
  _globals['_GRPC']._serialized_start=185
  _globals['_GRPC']._serialized_end=324
# @@protoc_insertion_point(module_scope)



from __future__ import annotations

import asyncio
import functools
import sys
import typing
from asyncio import Task

if sys.version_info >= (3, 10):  # pragma: no cover
    from typing import ParamSpec
else:  # pragma: no cover
    from typing_extensions import ParamSpec

from starlette.concurrency import run_in_threadpool

P = ParamSpec("P")


def is_async_callable(obj: typing.Any) -> typing.Any:
    while isinstance(obj, functools.partial):
        obj = obj.func

    return asyncio.iscoroutinefunction(obj) or (
            callable(obj) and asyncio.iscoroutinefunction(getattr(obj, "__call__"))
    )


class BackgroundTask:
    def __init__(
            self, func: typing.Callable[P, typing.Any], *args: P.args, **kwargs: P.kwargs
    ) -> None:
        self.func = func
        self.args = args
        self.kwargs = kwargs
        self.is_async = is_async_callable(func)

    async def __call__(self) -> None:
        if self.is_async:
            await self.func(*self.args, **self.kwargs)
        else:
            await run_in_threadpool(self.func, *self.args, **self.kwargs)


class BackgroundTasks:
    def __init__(self) -> None:
        self.queue = asyncio.Queue()
        self.is_running = False
        self.worker_task: typing.Union[Task, None] = None

    def add_task(
            self, func: typing.Callable[P, typing.Any], *args: P.args, **kwargs: P.kwargs
    ) -> None:
        task = BackgroundTask(func, *args, **kwargs)
        self.add(task)

    def add(self, task: BackgroundTask) -> None:
        self.queue.put_nowait(task)

    async def _worker(self) -> None:
        while self.is_running:
            task = await self.queue.get()
            try:
                await task()
            except Exception as e:
                print(f"Task failed: {e}")
            finally:
                self.queue.task_done()
            await asyncio.sleep(1)

    def run(self) -> None:
        if not self.is_running:
            self.is_running = True
            self.worker_task = asyncio.get_running_loop().create_task(self._worker())

    async def shutdown(self) -> None:
        self.is_running = False
        if self.worker_task:
            self.worker_task.cancel()
        await self.queue.join()



import asyncio
import traceback
from typing import Callable, Type, Union, TYPE_CHECKING

from rich.style import Style

from nestipy.common.logger import console
from nestipy.core import NestipyApplication, NestipyConfig, DiscoverService
from nestipy.microservice.client.base import MicroserviceOption
from nestipy.microservice.client.factory import ClientModuleFactory
from nestipy.microservice.proxy import MicroserviceProxy
from nestipy.microservice.server import MicroServiceServer

if TYPE_CHECKING:
    from nestipy.common.exception.interface import ExceptionFilter
    from nestipy.common.interceptor import NestipyInterceptor
    from nestipy.common.guards.can_activate import CanActivate


class NestipyMicroservice:
    app: NestipyApplication
    servers: list[MicroServiceServer] = []
    _ms_ready: bool = False
    coroutines: list = []

    def __init__(self, module: Type, option: list[MicroserviceOption], auto_init: bool = True):
        self.root_module = module
        self.option = option
        if auto_init:
            self.app = NestipyApplication()
            self.app.init(root_module=self.root_module)

    async def ms_setup(self):
        self.servers = []
        for opt in self.option:
            client_adapter = ClientModuleFactory.create(opt)
            server = MicroServiceServer(
                pub_sub=client_adapter,
            )
            self.servers.append(server)
        # self.app.add_module_root_module(MicroserviceServerModule, _init=True)
        # create all instance if not
        await self.app.setup()
        discover: DiscoverService = await self.app.get(DiscoverService)
        controllers = discover.get_all_controller()
        for server in self.servers:
            MicroserviceProxy(self.app, server).apply_routes(controllers)

        self._ms_ready = True

    async def start(self):
        self.coroutines = []
        # start server listener
        if not self._ms_ready:
            await self.ms_setup()
        for server in self.servers:
            self.coroutines.append(server.listen())
        await asyncio.gather(*self.coroutines)

    async def stop(self):
        # for task in self.coroutines:
        #     task.done()
        for server in self.servers:
            try:
                await server.close()
            except RuntimeError as e:
                traceback.print_exc()
                print(e)
            finally:
                pass

    def use_global_interceptors(self, *interceptors: Union[Type, "NestipyInterceptor"]):
        self.app.get_adapter().add_global_interceptors(*interceptors)
        self.app._add_root_module_provider(*interceptors)

    def use_global_filters(self, *filters: Union[Type, "ExceptionFilter"]):
        self.app.get_adapter().add_global_filters(*filters)
        self.app._add_root_module_provider(*filters)

    def use_global_guards(self, *guards: Union[Type, "CanActivate"]):
        self.app.get_adapter().add_global_guards(*guards)

    async def __call__(self, scope: dict, receive: Callable, send: Callable):
        if scope.get("type") == "lifespan":
            def create_task():
                asyncio.create_task(self.start())

            create_task()
            console.print(
                "[INFO]  Microservice startup completed",
                style=Style(color="green", bold=True),
            )


class NestipyConnectMicroservice(NestipyMicroservice, NestipyApplication):
    _running: bool = False

    def __init__(
            self, module: Type, config: NestipyConfig, option: list[MicroserviceOption]
    ):
        NestipyMicroservice.__init__(self, module, option, auto_init=False)
        NestipyApplication.__init__(self, config)
        self.app = self
        self.init(self.root_module)
        self.ms_task = None

    async def __call__(self, scope: dict, receive: Callable, send: Callable):
        await NestipyApplication.__call__(self, scope, receive, send)

    def start_all_microservices(self):
        def on_start():
            if self.ms_task is None:
                self.ms_task = asyncio.get_running_loop().create_task(self.start())
                console.print(
                    "INFO:     Microservice startup completed.",
                    style=Style(color="green", bold=True),
                )

        async def on_stop():
            if self.ms_task is not None:
                self.ms_task.cancel()
            await self.stop()

        self.app.on_startup(on_start)
        self.app.on_shutdown(on_stop)



class DataRequest:
    def __init__(self, topic: str, data: str):
        pass


class SubRequest:
    def __init__(self, topic: str):
        pass


class UnsubRequest:
    def __init__(self, topic: str):
        pass


class DataResponse:
    def __init__(self, topic: str, data: str):
        pass


class Empty:
    pass



import asyncio
import typing
from dataclasses import asdict
from typing import Any, AsyncIterator

from aiomqtt import Client

from .base import ClientProxy, MicroserviceOption, MqttClientOption


class MqttClientProxy(ClientProxy):
    client: Client

    def __init__(
            self,
            option: MicroserviceOption,
    ):
        super().__init__(option)
        _option = typing.cast(MqttClientOption, self.option.option or MqttClientOption())
        self.client = Client(**asdict(_option))

    async def slave(self) -> "ClientProxy":
        return MqttClientProxy(
            option=self.option,
        )

    async def connect(self):
        await self.client.__aenter__()

    async def _publish(self, topic, data):
        await self.client.publish(topic, data, retain=False)

    async def subscribe(self, *args, **kwargs):
        await self.client.subscribe(*args, **kwargs)

    async def unsubscribe(self, *args):
        await self.client.unsubscribe(*args)

    async def send_response(self, topic, data):
        await self._publish(topic, data)

    async def listen(self) -> AsyncIterator[str]:
        async for message in self.client.messages:
            yield message.payload.decode("utf-8")
        await asyncio.sleep(0.01)

    async def listen_response(self, from_topic: str, timeout: int = 30) -> Any:
        async for message in self.client.messages:
            return message.payload.decode("utf-8")
        await asyncio.sleep(0.01)

    async def close(self):
        await self.client.__aexit__(None, None, None)



from .base import ClientProxy, MicroserviceOption
from .grpc.client import GrpcClientProxy
from .mqtt import MqttClientProxy
from .nats import NatsClientProxy
from .option import RabbitMQClientOption, GrpcClientOption
from .option import TCPClientOption, RedisClientOption, NatsClientOption, MqttClientOption
from .option import Transport
from .rabbitmq import RabbitMQClientProxy
from .redis import RedisClientProxy
from .tcp import TCPClientProxy


class ClientModuleFactory:
    @classmethod
    def create(cls, option: MicroserviceOption) -> ClientProxy:
        match option.transport:
            case Transport.TCP:
                if option.option is None:
                    option.option = TCPClientOption()
                return TCPClientProxy(
                    option,
                )
            case Transport.REDIS:
                if option.option is None:
                    option.option = RedisClientOption()
                return RedisClientProxy(
                    option,
                )
            case Transport.MQTT:
                if option.option is None:
                    option.option = MqttClientOption()
                return MqttClientProxy(
                    option,
                )
            case Transport.RABBITMQ:
                if option.option is None:
                    option.option = RabbitMQClientOption()
                return RabbitMQClientProxy(
                    option,
                )
            case Transport.NATS:
                if option.option is None:
                    option.option = NatsClientOption()
                return NatsClientProxy(
                    option,
                )

            case Transport.GRPC:
                if option.option is None:
                    option.option = GrpcClientOption()
                return GrpcClientProxy(
                    option,
                )
            case Transport.CUSTOM:
                if option.proxy is not None and isinstance(option.proxy, ClientProxy):
                    return option.proxy
                else:
                    raise Exception("Unknown Transport")
            case _:
                raise Exception("Unknown Transport")



from abc import ABC, abstractmethod
from typing import Any


class BaseCommand(ABC):
    """Abstract base class for Nestipy commands."""

    _args: list = []
    _options: dict[str, Any] = {}

    @abstractmethod
    async def run(self):
        """The method to handle the command logic."""
        raise NotImplemented("Must be implement")

    def init(self, context: tuple):
        """Parse the combined context into arguments and options.

        Args:
            context (tuple): Combined arguments and options.

        """
        args = []
        options = {}

        it = iter(context)
        for item in it:
            if item.startswith('--'):
                key = item.lstrip('--')
                value = next(it, True)
                if value.startswith('--'):
                    value = True
                options[key] = value
            else:
                args.append(item)
        self._args = args
        self._options = options

    def get_opt(self, key: str = None, default=None):
        """Retrieve an option from the options dictionary safely."""
        return self._options.get(key, default) if key else self._options

    def get_arg(self, index: int = None, default=None):
        """Retrieve an argument by index from the arguments list safely."""
        try:
            return self._args[index] if index is not None else self._args
        except IndexError:
            return default



import asyncio

from nestipy.microservice.client import RedisClientProxy, MicroserviceOption
from nestipy.microservice.client.option import RedisClientOption, Transport


async def main():
    client = RedisClientProxy(
        MicroserviceOption(
            transport=Transport.REDIS,
            option=RedisClientOption(host="localhost", port=6379)
        ),
    )
    await client.send("test", "request")


if __name__ == "__main__":
    asyncio.run(main())
